@@ -5,7 +5,7 @@ use bytes::{Buf, Bytes};
 use log::debug;
 use pretty_hex::PrettyHex;
 use sdp::media_description::MediaDescription;
-use std::{convert::TryFrom, num::NonZeroU16};
+use std::{convert::TryFrom, net::IpAddr, num::NonZeroU16};
 use url::Url;
 
 use super::{Presentation, Stream};
@@ -184,11 +184,29 @@ static STATIC_PAYLOAD_TYPES: [Option<StaticPayloadType>; 35] = [
     }),
 ];
 
+/// Joins a control URL to a base URL in a non-RFC-compliant but common way.
+/// This matches what live555 and ffmpeg do.
+///
+/// See discussion at [#9](https://github.com/scottlamb/retina/issues/9).
 fn join_control(base_url: &Url, control: &str) -> Result<Url, String> {
     if control == "*" {
         return Ok(base_url.clone());
     }
-    base_url.join(control).map_err(|e| {
+    if let Ok(absolute_url) = Url::parse(control) {
+        return Ok(absolute_url);
+    }
+
+    Url::parse(&format!(
+        "{}{}{}",
+        base_url.as_str(),
+        if base_url.as_str().ends_with('/') {
+            ""
+        } else {
+            "/"
+        },
+        control
+    ))
+    .map_err(|e| {
         format!(
             "unable to join base url {} with control url {:?}: {}",
             base_url, control, e
@@ -206,11 +224,7 @@ pub(crate) fn get_cseq(response: &rtsp_types::Response<Bytes>) -> Option<u32> {
 /// Parses a [MediaDescription] to a [Stream].
 /// On failure, returns an error which is expected to be supplemented with
 /// the [MediaDescription] debug string and packed into a `RtspResponseError`.
-fn parse_media(
-    base_url: &Url,
-    alt_base_url: &Url,
-    media_description: &MediaDescription,
-) -> Result<Stream, String> {
+fn parse_media(base_url: &Url, media_description: &MediaDescription) -> Result<Stream, String> {
     let media = media_description.media_name.media.clone();
 
     // https://tools.ietf.org/html/rfc8866#section-5.14 says "If the <proto>
@@ -253,7 +267,6 @@ fn parse_media(
     let mut rtpmap = None;
     let mut fmtp = None;
     let mut control = None;
-    let mut alt_control = None;
     for a in &media_description.attributes {
         if a.key == "rtpmap" {
             let v = a
@@ -292,11 +305,6 @@ fn parse_media(
                 .as_deref()
                 .map(|c| join_control(base_url, c))
                 .transpose()?;
-            alt_control = a
-                .value
-                .as_deref()
-                .map(|c| join_control(alt_base_url, c))
-                .transpose()?;
         }
     }
 
@@ -357,7 +365,7 @@ fn parse_media(
         rtp_payload_type,
         depacketizer,
         control,
-        alt_control,
+        sockets: None,
         channels,
         state: super::StreamState::Uninit,
     })
@@ -406,9 +414,7 @@ pub(crate) fn parse_describe(
                 .map(|v| (rtsp_types::headers::CONTENT_LOCATION, v))
         })
         .map(|(h, v)| Url::parse(v.as_str()).map_err(|e| format!("bad {} {:?}: {}", h, v, e)))
-        .unwrap_or(Ok(request_url))?;
-    let mut alt_base_url = base_url.clone();
-    alt_base_url.set_path(&format!("{}/", base_url.path()));
+        .unwrap_or(Ok(request_url.clone()))?;
 
     let mut control = None;
     for a in &sdp.attributes {
@@ -421,14 +427,14 @@ pub(crate) fn parse_describe(
             break;
         }
     }
-    let control = control.ok_or_else(|| "no control url".to_string())?;
+    let control = control.unwrap_or(request_url);
 
     let streams = sdp
         .media_descriptions
         .iter()
         .enumerate()
         .map(|(i, m)| {
-            parse_media(&base_url, &alt_base_url, &m)
+            parse_media(&base_url, &m)
                 .map_err(|e| format!("Unable to parse stream {}: {}\n\n{:#?}", i, &e, &m))
         })
         .collect::<Result<Vec<Stream>, String>>()?;
@@ -441,14 +447,15 @@ pub(crate) fn parse_describe(
         base_url,
         control,
         accept_dynamic_rate,
-        sdp,
     })
 }
 
 pub(crate) struct SetupResponse<'a> {
     pub(crate) session_id: &'a str,
     pub(crate) ssrc: Option<u32>,
-    pub(crate) channel_id: u8,
+    pub(crate) channel_id: Option<u8>,
+    pub(crate) source: Option<IpAddr>,
+    pub(crate) server_port: Option<(u16, u16)>,
 }
 
 /// Parses a `SETUP` response.
@@ -468,6 +475,8 @@ pub(crate) fn parse_setup(response: &rtsp_types::Response<Bytes>) -> Result<Setu
         .ok_or_else(|| "Missing Transport header".to_string())?;
     let mut channel_id = None;
     let mut ssrc = None;
+    let mut source = None;
+    let mut server_port = None;
     for part in transport.as_str().split(';') {
         if let Some(v) = part.strip_prefix("ssrc=") {
             let v = u32::from_str_radix(v, 16).map_err(|_| format!("Unparseable ssrc {}", v))?;
@@ -485,14 +494,33 @@ pub(crate) fn parse_setup(response: &rtsp_types::Response<Bytes>) -> Result<Setu
                 }
             }
             channel_id = Some(n);
+        } else if let Some(s) = part.strip_prefix("source=") {
+            source = Some(
+                s.parse()
+                    .map_err(|_| format!("Transport header has unparseable source {:?}", s))?,
+            );
+        } else if let Some(s) = part.strip_prefix("server_port=") {
+            let mut ports = s.splitn(2, '-');
+            let n = ports.next().expect("splitn returns at least one part");
+            let n = u16::from_str_radix(n, 10)
+                .map_err(|_| format!("bad port in Transport: {}", transport.as_str()))?;
+            if let Some(m) = ports.next() {
+                let m = u16::from_str_radix(m, 10).map_err(|_| format!("bad second port {}", m))?;
+                server_port = Some((n, m))
+            } else {
+                // TODO: this is allowed by RFC 2326's grammar, but I'm not sure
+                // what it means. Does it use the same port for both RTP and
+                // RTCP, or is it implied the second is one more than the first?
+                return Err("Transport header specifies a single server_port".to_owned());
+            }
         }
     }
-    let channel_id =
-        channel_id.ok_or_else(|| "Transport header has no interleaved parameter".to_string())?;
     Ok(SetupResponse {
         session_id,
         ssrc,
         channel_id,
+        source,
+        server_port,
     })
 }
 
@@ -502,9 +530,10 @@ pub(crate) fn parse_play(
     presentation: &mut Presentation,
 ) -> Result<(), String> {
     // https://tools.ietf.org/html/rfc2326#section-12.33
-    let rtp_info = response
-        .header(&rtsp_types::headers::RTP_INFO)
-        .ok_or_else(|| "PLAY response has no RTP-Info header".to_string())?;
+    let rtp_info = match response.header(&rtsp_types::headers::RTP_INFO) {
+        Some(rtsp_info) => rtsp_info,
+        None => return Ok(()),
+    };
     for s in rtp_info.as_str().split(',') {
         let s = s.trim();
         let mut parts = s.split(';');
@@ -514,7 +543,7 @@ pub(crate) fn parse_play(
             .strip_prefix("url=")
             .ok_or_else(|| "RTP-Info missing stream URL".to_string())?;
         let url = join_control(&presentation.base_url, url)?;
-        let mut stream;
+        let stream;
         if presentation.streams.len() == 1 {
             // The server is allowed to not specify a stream control URL for
             // single-stream presentations. Additionally, some buggy
@@ -530,18 +559,14 @@ pub(crate) fn parse_play(
                 .streams
                 .iter_mut()
                 .find(|s| matches!(&s.control, Some(u) if u == &url));
-
-            // If we didn't find a stream, try again with alt_control. Don't do
-            // this on the first pass because we should check all of the
-            // proper control URLs first.
-            if stream.is_none() {
-                stream = presentation
-                    .streams
-                    .iter_mut()
-                    .find(|s| matches!(&s.alt_control, Some(u) if u == &url));
-            }
         }
-        let stream = stream.ok_or_else(|| format!("RTP-Info contains unknown stream {}", url))?;
+        let stream = match stream {
+            Some(s) => s,
+            None => {
+                log::warn!("RTP-Info contains unknown stream {}", url);
+                continue;
+            }
+        };
         let state = match &mut stream.state {
             super::StreamState::Uninit => {
                 // This appears to happen for Reolink devices when we did not send a SETUP request
@@ -587,21 +612,12 @@ pub(crate) fn parse_play(
 mod tests {
     use std::num::NonZeroU16;
 
-    use bytes::Bytes;
     use url::Url;
 
     use crate::{client::StreamStateInit, codec::Parameters};
 
     use super::super::StreamState;
-
-    fn response(raw: &'static [u8]) -> rtsp_types::Response<Bytes> {
-        let (msg, len) = rtsp_types::Message::parse(raw).unwrap();
-        assert_eq!(len, raw.len());
-        match msg {
-            rtsp_types::Message::Response(r) => r.map_body(|b| Bytes::from_static(b)),
-            _ => panic!("unexpected message type"),
-        }
-    }
+    use crate::testutil::response;
 
     fn parse_describe(
         raw_url: &str,
@@ -614,16 +630,14 @@ mod tests {
     #[test]
     fn dahua_h264_aac_onvif() {
         // DESCRIBE.
-        let prefix = "rtsp://192.168.5.111:554/cam/";
+        let prefix =
+            "rtsp://192.168.5.111:554/cam/realmonitor?channel=1&subtype=1&unicast=true&proto=Onvif";
         let mut p = parse_describe(
-            &(prefix.to_string() + "realmonitor?channel=1&subtype=1&unicast=true&proto=Onvif"),
+            prefix,
             include_bytes!("testdata/dahua_describe_h264_aac_onvif.txt"),
         )
         .unwrap();
-        assert_eq!(
-            p.control.as_str(),
-            &(prefix.to_string() + "realmonitor?channel=1&subtype=1&unicast=true&proto=Onvif/")
-        );
+        assert_eq!(p.control.as_str(), &(prefix.to_string() + "/"));
         assert!(p.accept_dynamic_rate);
 
         assert_eq!(p.streams.len(), 3);
@@ -631,7 +645,7 @@ mod tests {
         // H.264 video stream.
         assert_eq!(
             p.streams[0].control.as_ref().unwrap().as_str(),
-            &(prefix.to_string() + "trackID=0")
+            &(prefix.to_string() + "/trackID=0")
         );
         assert_eq!(p.streams[0].media, "video");
         assert_eq!(p.streams[0].encoding_name, "h264");
@@ -650,7 +664,7 @@ mod tests {
         // .mp4 audio stream.
         assert_eq!(
             p.streams[1].control.as_ref().unwrap().as_str(),
-            &(prefix.to_string() + "trackID=1")
+            &(prefix.to_string() + "/trackID=1")
         );
         assert_eq!(p.streams[1].media, "audio");
         assert_eq!(p.streams[1].encoding_name, "mpeg4-generic");
@@ -664,7 +678,7 @@ mod tests {
         // ONVIF parameters stream.
         assert_eq!(
             p.streams[2].control.as_ref().unwrap().as_str(),
-            &(prefix.to_string() + "trackID=4")
+            &(prefix.to_string() + "/trackID=4")
         );
         assert_eq!(p.streams[2].media, "application");
         assert_eq!(p.streams[2].encoding_name, "vnd.onvif.metadata");
@@ -679,7 +693,7 @@ mod tests {
         let setup_response = response(include_bytes!("testdata/dahua_setup.txt"));
         let setup_response = super::parse_setup(&setup_response).unwrap();
         assert_eq!(setup_response.session_id, "634214675641");
-        assert_eq!(setup_response.channel_id, 0);
+        assert_eq!(setup_response.channel_id, Some(0));
         assert_eq!(setup_response.ssrc, Some(0x30a98ee7));
         p.streams[0].state = StreamState::Init(StreamStateInit {
             ssrc: setup_response.ssrc,
@@ -776,7 +790,7 @@ mod tests {
         let setup_response = response(include_bytes!("testdata/hikvision_setup.txt"));
         let setup_response = super::parse_setup(&setup_response).unwrap();
         assert_eq!(setup_response.session_id, "708345999");
-        assert_eq!(setup_response.channel_id, 0);
+        assert_eq!(setup_response.channel_id, Some(0));
         assert_eq!(setup_response.ssrc, Some(0x4cacc3d1));
         p.streams[0].state = StreamState::Init(StreamStateInit {
             ssrc: setup_response.ssrc,
@@ -851,7 +865,7 @@ mod tests {
         let setup_response = response(include_bytes!("testdata/reolink_setup.txt"));
         let setup_response = super::parse_setup(&setup_response).unwrap();
         assert_eq!(setup_response.session_id, "F8F8E425");
-        assert_eq!(setup_response.channel_id, 0);
+        assert_eq!(setup_response.channel_id, Some(0));
         assert_eq!(setup_response.ssrc, None);
         p.streams[0].state = StreamState::Init(StreamStateInit::default());
         p.streams[1].state = StreamState::Init(StreamStateInit::default());
@@ -929,7 +943,7 @@ mod tests {
         let setup_response = response(include_bytes!("testdata/bunny_setup.txt"));
         let setup_response = super::parse_setup(&setup_response).unwrap();
         assert_eq!(setup_response.session_id, "1642021126");
-        assert_eq!(setup_response.channel_id, 0);
+        assert_eq!(setup_response.channel_id, Some(0));
         assert_eq!(setup_response.ssrc, None);
         p.streams[0].state = StreamState::Init(StreamStateInit::default());
         p.streams[1].state = StreamState::Init(StreamStateInit::default());
@@ -991,6 +1005,55 @@ mod tests {
         };
     }
 
+    #[test]
+    fn vstarcam() {
+        // DESCRIBE.
+        let prefix = "rtsp://192.168.1.198:10554/tcp/av0_0";
+        let p = parse_describe(prefix, include_bytes!("testdata/vstarcam_describe.txt")).unwrap();
+        assert_eq!(p.control.as_str(), &(prefix.to_string()));
+
+        assert!(!p.accept_dynamic_rate);
+
+        assert_eq!(p.streams.len(), 2);
+
+        // H.264 video stream.
+        assert_eq!(
+            p.streams[0].control.as_ref().unwrap().as_str(),
+            &(prefix.to_string() + "/track0")
+        );
+
+        assert_eq!(p.streams[0].media, "video");
+
+        assert_eq!(p.streams[0].encoding_name, "h264");
+        assert_eq!(p.streams[0].rtp_payload_type, 96);
+        assert_eq!(p.streams[0].clock_rate, 90_000);
+
+        match p.streams[0].parameters().unwrap() {
+            Parameters::Video(v) => {
+                assert_eq!(v.rfc6381_codec(), "avc1.4D002A");
+                assert_eq!(v.pixel_dimensions(), (1920, 1080));
+                assert_eq!(v.pixel_aspect_ratio(), None);
+                assert_eq!(v.frame_rate(), Some((2, 15)));
+            }
+            _ => panic!(),
+        }
+
+        // audio stream
+        assert_eq!(
+            p.streams[1].control.as_ref().unwrap().as_str(),
+            &(prefix.to_string() + "/track1")
+        );
+        assert_eq!(p.streams[1].media, "audio");
+        assert_eq!(p.streams[1].encoding_name, "pcma");
+        assert_eq!(p.streams[1].rtp_payload_type, 8);
+        assert_eq!(p.streams[1].clock_rate, 8_000);
+        assert_eq!(p.streams[1].channels, NonZeroU16::new(1));
+        match p.streams[1].parameters().unwrap() {
+            Parameters::Audio(_) => {}
+            _ => panic!(),
+        };
+    }
+
     /// [GW Security GW4089IP](https://github.com/scottlamb/moonfire-nvr/wiki/Cameras:-GW-Security#gw4089ip),
     /// main stream (high-res, audio).
     #[test]
@@ -1006,7 +1069,7 @@ mod tests {
         // H.264 video stream.
         assert_eq!(
             p.streams[0].control.as_ref().unwrap().as_str(),
-            "rtsp://192.168.1.110:5050/video"
+            "rtsp://192.168.1.110:5050/H264?channel=1&subtype=0&unicast=true&proto=Onvif/video"
         );
         assert_eq!(p.streams[0].media, "video");
         assert_eq!(p.streams[0].encoding_name, "h264");
@@ -1025,7 +1088,7 @@ mod tests {
         // audio stream
         assert_eq!(
             p.streams[1].control.as_ref().unwrap().as_str(),
-            "rtsp://192.168.1.110:5050/audio"
+            "rtsp://192.168.1.110:5050/H264?channel=1&subtype=0&unicast=true&proto=Onvif/audio"
         );
         assert_eq!(p.streams[1].media, "audio");
         assert_eq!(p.streams[1].encoding_name, "pcmu"); // rtpmap wins over static list.
@@ -1041,7 +1104,7 @@ mod tests {
         let setup_response = response(include_bytes!("testdata/gw_main_setup_video.txt"));
         let setup_response = super::parse_setup(&setup_response).unwrap();
         assert_eq!(setup_response.session_id, "9a90de54");
-        assert_eq!(setup_response.channel_id, 0);
+        assert_eq!(setup_response.channel_id, Some(0));
         assert_eq!(setup_response.ssrc, None);
         p.streams[0].state = StreamState::Init(StreamStateInit {
             ssrc: None,
@@ -1052,7 +1115,7 @@ mod tests {
         let setup_response = response(include_bytes!("testdata/gw_main_setup_audio.txt"));
         let setup_response = super::parse_setup(&setup_response).unwrap();
         assert_eq!(setup_response.session_id, "9a90de54");
-        assert_eq!(setup_response.channel_id, 2);
+        assert_eq!(setup_response.channel_id, Some(2));
         assert_eq!(setup_response.ssrc, None);
         p.streams[1].state = StreamState::Init(StreamStateInit {
             ssrc: None,
@@ -1066,10 +1129,11 @@ mod tests {
             &mut p,
         )
         .unwrap();
+        // The RTP-Info url= isn't in the expected format, so its contents are skipped.
         match &p.streams[0].state {
             StreamState::Init(s) => {
-                assert_eq!(s.initial_seq, Some(271));
-                assert_eq!(s.initial_rtptime, Some(1621990950));
+                assert_eq!(s.initial_seq, None);
+                assert_eq!(s.initial_rtptime, None);
             }
             _ => panic!(),
         };
@@ -1089,7 +1153,10 @@ mod tests {
         // DESCRIBE.
         let base = "rtsp://192.168.1.110:5049/H264?channel=1&subtype=1&unicast=true&proto=Onvif";
         let mut p = parse_describe(base, include_bytes!("testdata/gw_sub_describe.txt")).unwrap();
-        assert_eq!(p.control.as_str(), base);
+        assert_eq!(
+            p.control.as_str(),
+            "rtsp://192.168.1.110:5049/H264?channel=1&subtype=1&unicast=true&proto=Onvif"
+        );
         assert!(!p.accept_dynamic_rate);
 
         assert_eq!(p.streams.len(), 1);
@@ -1097,7 +1164,7 @@ mod tests {
         // H.264 video stream.
         assert_eq!(
             p.streams[0].control.as_ref().unwrap().as_str(),
-            "rtsp://192.168.1.110:5049/video"
+            "rtsp://192.168.1.110:5049/H264?channel=1&subtype=1&unicast=true&proto=Onvif/video"
         );
         assert_eq!(p.streams[0].media, "video");
         assert_eq!(p.streams[0].encoding_name, "h264");
@@ -1117,7 +1184,7 @@ mod tests {
         let setup_response = response(include_bytes!("testdata/gw_sub_setup.txt"));
         let setup_response = super::parse_setup(&setup_response).unwrap();
         assert_eq!(setup_response.session_id, "9b0d0e54");
-        assert_eq!(setup_response.channel_id, 0);
+        assert_eq!(setup_response.channel_id, Some(0));
         assert_eq!(setup_response.ssrc, None);
         p.streams[0].state = StreamState::Init(StreamStateInit {
             ssrc: None,
