@@ -12,8 +12,7 @@ use crate::{Error, ErrorInt};
 
 /// A received RTP packet.
 pub struct Packet {
-    pub ctx: crate::RtspMessageContext,
-    pub channel_id: u8,
+    pub ctx: crate::PacketContext,
     pub stream_id: usize,
     pub timestamp: crate::Timestamp,
     pub ssrc: u32,
@@ -37,7 +36,6 @@ impl std::fmt::Debug for Packet {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         f.debug_struct("Packet")
             .field("ctx", &self.ctx)
-            .field("channel_id", &self.channel_id)
             .field("stream_id", &self.stream_id)
             .field("timestamp", &self.timestamp)
             .field("ssrc", &self.ssrc)
@@ -53,7 +51,7 @@ impl std::fmt::Debug for Packet {
 #[derive(Debug)]
 pub struct SenderReport {
     pub stream_id: usize,
-    pub ctx: crate::RtspMessageContext,
+    pub ctx: crate::PacketContext,
     pub timestamp: crate::Timestamp,
     pub ntp_timestamp: crate::NtpTimestamp,
 }
@@ -61,13 +59,16 @@ pub struct SenderReport {
 /// RTP/RTCP demarshaller which ensures packets have the correct SSRC and
 /// monotonically increasing SEQ. Unstable; exposed for benchmark.
 ///
-/// This reports packet loss (via [Packet::loss]) but doesn't prohibit it, except for losses
+/// When using UDP, skips and logs out-of-order packets. When using TCP,
+/// fails on them.
+///
+/// This reports packet loss (via [Packet::loss]) but doesn't prohibit it
 /// of more than `i16::MAX` which would be indistinguishable from non-monotonic sequence numbers.
 /// Servers sometimes drop packets internally even when sending data via TCP.
 ///
 /// At least [one camera](https://github.com/scottlamb/moonfire-nvr/wiki/Cameras:-Reolink#reolink-rlc-410-hardware-version-ipc_3816m)
 /// sometimes sends data from old RTSP sessions over new ones. This seems like a
-/// serious bug, and currently `StrictSequenceChecker` will error in this case,
+/// serious bug, and currently `InorderRtpParser` will error in this case,
 /// although it'd be possible to discard the incorrect SSRC instead.
 ///
 /// [RFC 3550 section 8.2](https://tools.ietf.org/html/rfc3550#section-8.2) says that SSRC
@@ -75,43 +76,29 @@ pub struct SenderReport {
 /// not sure it will ever come up with IP cameras.
 #[doc(hidden)]
 #[derive(Debug)]
-pub struct StrictSequenceChecker {
+pub struct InorderParser {
     ssrc: Option<u32>,
     next_seq: Option<u16>,
 }
 
-impl StrictSequenceChecker {
+impl InorderParser {
     pub fn new(ssrc: Option<u32>, next_seq: Option<u16>) -> Self {
         Self { ssrc, next_seq }
     }
 
     pub fn rtp(
         &mut self,
+        session_options: &super::SessionOptions,
         conn_ctx: &crate::ConnectionContext,
-        msg_ctx: &crate::RtspMessageContext,
+        pkt_ctx: &crate::PacketContext,
         timeline: &mut super::Timeline,
-        channel_id: u8,
         stream_id: usize,
         mut data: Bytes,
-    ) -> Result<PacketItem, Error> {
-        // Terrible hack to try to make sense of the GW Security GW4089IP's audio stream.
-        // It appears to have one RTSP interleaved message wrapped in another. RTP and RTCP
-        // packets can never start with '$', so this shouldn't interfere with well-behaved
-        // servers.
-        if data.len() > 4
-            && data[0] == b'$'
-            && usize::from(u16::from_be_bytes([data[2], data[3]])) <= data.len() - 4
-        {
-            log::debug!("stripping extra interleaved data header");
-            data.advance(4);
-            // also remove suffix? dunno.
-        }
-
+    ) -> Result<Option<PacketItem>, Error> {
         let reader = rtp_rs::RtpReader::new(&data[..]).map_err(|e| {
-            wrap!(ErrorInt::RtspDataMessageError {
+            wrap!(ErrorInt::PacketError {
                 conn_ctx: *conn_ctx,
-                msg_ctx: *msg_ctx,
-                channel_id,
+                pkt_ctx: *pkt_ctx,
                 stream_id,
                 description: format!(
                     "corrupt RTP header while expecting seq={:04x?}: {:?}\n{:#?}",
@@ -123,40 +110,59 @@ impl StrictSequenceChecker {
         })?;
         let sequence_number = u16::from_be_bytes([data[2], data[3]]); // I don't like rtsp_rs::Seq.
         let ssrc = reader.ssrc();
+        let loss = sequence_number.wrapping_sub(self.next_seq.unwrap_or(sequence_number));
+        if matches!(self.ssrc, Some(s) if s != ssrc) {
+            bail!(ErrorInt::RtpPacketError {
+                conn_ctx: *conn_ctx,
+                pkt_ctx: *pkt_ctx,
+                stream_id,
+                ssrc,
+                sequence_number,
+                description: format!(
+                    "Wrong ssrc; expecting ssrc={:08x?} seq={:04x?}",
+                    self.ssrc, self.next_seq
+                ),
+            });
+        }
+        if loss > 0x80_00 {
+            if matches!(session_options.transport, super::Transport::Tcp) {
+                bail!(ErrorInt::RtpPacketError {
+                    conn_ctx: *conn_ctx,
+                    pkt_ctx: *pkt_ctx,
+                    stream_id,
+                    ssrc,
+                    sequence_number,
+                    description: format!(
+                        "Out-of-order packet or large loss; expecting ssrc={:08x?} seq={:04x?}",
+                        self.ssrc, self.next_seq
+                    ),
+                });
+            } else {
+                log::info!(
+                    "Skipping out-of-order seq={:04x} when expecting ssrc={:08x?} seq={:04x?}",
+                    sequence_number,
+                    self.ssrc,
+                    self.next_seq
+                );
+            }
+        }
         let timestamp = match timeline.advance_to(reader.timestamp()) {
             Ok(ts) => ts,
             Err(description) => bail!(ErrorInt::RtpPacketError {
                 conn_ctx: *conn_ctx,
-                msg_ctx: *msg_ctx,
-                channel_id,
+                pkt_ctx: *pkt_ctx,
                 stream_id,
                 ssrc,
                 sequence_number,
                 description,
             }),
         };
-        let loss = sequence_number.wrapping_sub(self.next_seq.unwrap_or(sequence_number));
-        if matches!(self.ssrc, Some(s) if s != ssrc) || loss > 0x80_00 {
-            bail!(ErrorInt::RtpPacketError {
-                conn_ctx: *conn_ctx,
-                msg_ctx: *msg_ctx,
-                channel_id,
-                stream_id,
-                ssrc,
-                sequence_number,
-                description: format!(
-                    "Expected ssrc={:08x?} seq={:04x?}",
-                    self.ssrc, self.next_seq
-                ),
-            });
-        }
         self.ssrc = Some(ssrc);
         let mark = reader.mark();
         let payload_range = crate::as_range(&data, reader.payload()).ok_or_else(|| {
             wrap!(ErrorInt::RtpPacketError {
                 conn_ctx: *conn_ctx,
-                msg_ctx: *msg_ctx,
-                channel_id,
+                pkt_ctx: *pkt_ctx,
                 stream_id,
                 ssrc,
                 sequence_number,
@@ -166,9 +172,8 @@ impl StrictSequenceChecker {
         data.truncate(payload_range.end);
         data.advance(payload_range.start);
         self.next_seq = Some(sequence_number.wrapping_add(1));
-        Ok(PacketItem::RtpPacket(Packet {
-            ctx: *msg_ctx,
-            channel_id,
+        Ok(Some(PacketItem::RtpPacket(Packet {
+            ctx: *pkt_ctx,
             stream_id,
             timestamp,
             ssrc,
@@ -176,12 +181,12 @@ impl StrictSequenceChecker {
             loss,
             mark,
             payload: data,
-        }))
+        })))
     }
 
     pub fn rtcp(
         &mut self,
-        msg_ctx: &crate::RtspMessageContext,
+        pkt_ctx: &crate::PacketContext,
         timeline: &mut super::Timeline,
         stream_id: usize,
         data: Bytes,
@@ -216,12 +221,12 @@ impl StrictSequenceChecker {
 
                     sr = Some(SenderReport {
                         stream_id,
-                        ctx: *msg_ctx,
+                        ctx: *pkt_ctx,
                         timestamp,
                         ntp_timestamp: pkt.ntp_timestamp(),
                     });
                 }
-                crate::rtcp::Packet::Unknown(pkt) => debug!("rtcp: {:?}", pkt.payload_type()),
+                crate::rtcp::Packet::Unknown(pkt) => debug!("rtcp: pt {:?}", pkt.payload_type()),
             }
             i += 1;
         }
