@@ -4,10 +4,11 @@
 //! [H.264](https://www.itu.int/rec/T-REC-H.264-201906-I/en)-encoded video.
 
 use std::convert::TryFrom;
+use std::fmt::Write;
 
 use bytes::{Buf, BufMut, Bytes, BytesMut};
 use h264_reader::nal::{NalHeader, UnitType};
-use log::debug;
+use log::{debug, log_enabled, trace};
 
 use crate::{client::rtp::Packet, Error, Timestamp};
 
@@ -55,8 +56,8 @@ struct Nal {
 /// An access unit that is currently being accumulated during `PreMark` state.
 #[derive(Debug)]
 struct AccessUnit {
-    start_ctx: crate::RtspMessageContext,
-    end_ctx: crate::RtspMessageContext,
+    start_ctx: crate::PacketContext,
+    end_ctx: crate::PacketContext,
     timestamp: crate::Timestamp,
     stream_id: usize,
 
@@ -65,6 +66,8 @@ struct AccessUnit {
 
     /// RTP packets lost as this access unit was starting.
     loss: u16,
+
+    same_ts_as_prev: bool,
 }
 
 #[derive(Debug)]
@@ -131,7 +134,7 @@ impl Depacketizer {
                 DepacketizerInputState::New => {
                     debug_assert!(self.nals.is_empty());
                     debug_assert!(self.pieces.is_empty());
-                    AccessUnit::start(&pkt, 0)
+                    AccessUnit::start(&pkt, 0, false)
                 }
                 DepacketizerInputState::PreMark(mut access_unit) => {
                     if pkt.loss > 0 {
@@ -156,7 +159,7 @@ impl Depacketizer {
                         }
                         // A suffix of a previous access unit was lost; discard it.
                         // A prefix of the new one may have been lost; try parsing.
-                        AccessUnit::start(&pkt, 0)
+                        AccessUnit::start(&pkt, 0, false)
                     } else if access_unit.timestamp.timestamp != pkt.timestamp.timestamp {
                         if access_unit.in_fu_a {
                             return Err(format!(
@@ -164,9 +167,20 @@ impl Depacketizer {
                                 access_unit.timestamp, pkt.timestamp
                             ));
                         }
-                        access_unit.end_ctx = pkt.ctx;
-                        self.pending = Some(self.finalize_access_unit(access_unit)?);
-                        AccessUnit::start(&pkt, 0)
+                        let last_nal_hdr = self.nals.last().unwrap().hdr;
+                        if can_end_au(last_nal_hdr.nal_unit_type()) {
+                            access_unit.end_ctx = pkt.ctx;
+                            self.pending =
+                                Some(self.finalize_access_unit(access_unit, "ts change")?);
+                            AccessUnit::start(&pkt, 0, false)
+                        } else {
+                            log::debug!(
+                                "Bogus mid-access unit timestamp change after {:?}",
+                                last_nal_hdr
+                            );
+                            access_unit.timestamp.timestamp = pkt.timestamp.timestamp;
+                            access_unit
+                        }
                     } else {
                         access_unit
                     }
@@ -177,10 +191,7 @@ impl Depacketizer {
                 } => {
                     debug_assert!(self.nals.is_empty());
                     debug_assert!(self.pieces.is_empty());
-                    if state_ts.timestamp == pkt.timestamp.timestamp {
-                        return Err("packet follows marked packet with same timestamp".into());
-                    }
-                    AccessUnit::start(&pkt, loss)
+                    AccessUnit::start(&pkt, loss, state_ts.timestamp == pkt.timestamp.timestamp)
                 }
                 DepacketizerInputState::Loss {
                     timestamp,
@@ -193,7 +204,7 @@ impl Depacketizer {
                         self.input_state = DepacketizerInputState::Loss { timestamp, pkts };
                         return Ok(());
                     }
-                    AccessUnit::start(&pkt, pkts)
+                    AccessUnit::start(&pkt, pkts, false)
                 }
             };
 
@@ -340,11 +351,21 @@ impl Depacketizer {
             _ => return Err(format!("bad nal header {:02x}", nal_header)),
         }
         self.input_state = if pkt.mark {
-            access_unit.end_ctx = pkt.ctx;
-            self.pending = Some(self.finalize_access_unit(access_unit)?);
-            DepacketizerInputState::PostMark {
-                timestamp: pkt.timestamp,
-                loss: 0,
+            let last_nal_hdr = self.nals.last().unwrap().hdr;
+            if can_end_au(last_nal_hdr.nal_unit_type()) {
+                access_unit.end_ctx = pkt.ctx;
+                self.pending = Some(self.finalize_access_unit(access_unit, "mark")?);
+                DepacketizerInputState::PostMark {
+                    timestamp: pkt.timestamp,
+                    loss: 0,
+                }
+            } else {
+                log::debug!(
+                    "Bogus mid-access unit timestamp change after {:?}",
+                    last_nal_hdr
+                );
+                access_unit.timestamp.timestamp = pkt.timestamp.timestamp;
+                DepacketizerInputState::PreMark(access_unit)
             }
         } else {
             DepacketizerInputState::PreMark(access_unit)
@@ -362,13 +383,49 @@ impl Depacketizer {
         u32::try_from(self.pieces.len()).map_err(|_| "more than u32::MAX pieces!".to_string())
     }
 
-    fn finalize_access_unit(&mut self, au: AccessUnit) -> Result<VideoFrame, String> {
+    /// Logs information about each access unit.
+    /// Currently, "bad" access units (violating certain specification rules)
+    /// are logged at debug priority, and others are logged at trace priority.
+    fn log_access_unit(&self, au: &AccessUnit, reason: &str) {
+        let mut errs = String::new();
+        if au.same_ts_as_prev {
+            errs.push_str("\n* same timestamp as previous access unit");
+        }
+        validate_order(&self.nals, &mut errs);
+        if !errs.is_empty() {
+            let mut nals = String::new();
+            for (i, nal) in self.nals.iter().enumerate() {
+                let _ = write!(&mut nals, "\n  {}: {:?}", i, nal.hdr);
+            }
+            debug!(
+                "bad access unit (ended by {}) at ts {}\nerrors are:{}\nNALs are:{}",
+                reason, au.timestamp, errs, nals
+            );
+        } else if log_enabled!(log::Level::Trace) {
+            let mut nals = String::new();
+            for (i, nal) in self.nals.iter().enumerate() {
+                let _ = write!(&mut nals, "\n  {}: {:?}", i, nal.hdr);
+            }
+            trace!(
+                "access unit (ended by {}) at ts {}; NALS are:{}",
+                reason,
+                au.timestamp,
+                nals
+            );
+        }
+    }
+
+    fn finalize_access_unit(&mut self, au: AccessUnit, reason: &str) -> Result<VideoFrame, String> {
         let mut piece_idx = 0;
         let mut retained_len = 0usize;
         let mut is_random_access_point = false;
         let mut is_disposable = true;
         let mut new_sps = None;
         let mut new_pps = None;
+
+        if log_enabled!(log::Level::Debug) {
+            self.log_access_unit(&au, reason);
+        }
         for nal in &self.nals {
             let next_piece_idx = usize::try_from(nal.next_piece_idx).expect("u32 fits in usize");
             let nal_pieces = &self.pieces[piece_idx..next_piece_idx];
@@ -439,8 +496,26 @@ impl Depacketizer {
     }
 }
 
+/// Returns true if we allow the given NAL unit type to end an access unit.
+///
+/// We specifically prohibit this for the SPS and PPS. Reolink cameras sometimes
+/// incorrectly set the RTP marker bit and/or change the timestamp after these.
+fn can_end_au(nal_unit_type: UnitType) -> bool {
+    // H.264 section 7.4.1.2.3 Order of NAL units and coded pictures and
+    // association to access units says "Sequence parameter set NAL units or
+    // picture parameter set NAL units may be present in an access unit, but
+    // cannot follow the last VCL NAL unit of the primary coded picture within
+    // the access unit, as this condition would specify the start of a new
+    // access unit."
+    nal_unit_type != UnitType::SeqParameterSet && nal_unit_type != UnitType::PicParameterSet
+}
+
 impl AccessUnit {
-    fn start(pkt: &crate::client::rtp::Packet, additional_loss: u16) -> Self {
+    fn start(
+        pkt: &crate::client::rtp::Packet,
+        additional_loss: u16,
+        same_ts_as_prev: bool,
+    ) -> Self {
         AccessUnit {
             start_ctx: pkt.ctx,
             end_ctx: pkt.ctx,
@@ -450,10 +525,55 @@ impl AccessUnit {
 
             // TODO: overflow?
             loss: pkt.loss + additional_loss,
+            same_ts_as_prev,
         }
     }
 }
 
+/// Checks NAL unit type ordering against rules of H.264 section 7.4.1.2.3.
+///
+/// This doesn't precisely check every rule there but enough to diagnose some
+/// problems.
+fn validate_order(nals: &[Nal], errs: &mut String) {
+    let mut seen_vcl = false;
+    for (i, nal) in nals.iter().enumerate() {
+        match nal.hdr.nal_unit_type() {
+            /* 1 */ UnitType::SliceLayerWithoutPartitioningNonIdr |
+            /* 2 */ UnitType::SliceDataPartitionALayer |
+            /* 3 */ UnitType::SliceDataPartitionBLayer |
+            /* 4 */ UnitType::SliceDataPartitionCLayer |
+            /* 5 */ UnitType::SliceLayerWithoutPartitioningIdr => {
+                seen_vcl = true;
+            }
+            /* 6 */ UnitType::SEI => {
+                if seen_vcl {
+                    errs.push_str("\n* SEI after VCL");
+                }
+            }
+            /* 9 */ UnitType::AccessUnitDelimiter => {
+                if i != 0 {
+                    let _ = write!(errs, "access unit delimiter must be first in AU; was preceded by {:?}",
+                                nals[i-1].hdr);
+                }
+            }
+            /* 10 */ UnitType::EndOfSeq => {
+                if !seen_vcl {
+                    errs.push_str("\n* end of sequence without VCL");
+                }
+            }
+            /* 11 */ UnitType::EndOfStream => {
+                if i != nals.len() - 1 {
+                    errs.push_str("\n* end of stream NAL isn't last");
+                }
+            }
+            _ => {}
+        }
+    }
+    if !seen_vcl {
+        errs.push_str("\n* missing VCL");
+    }
+}
+
 #[derive(Clone, Debug)]
 struct InternalParameters {
     generic_parameters: super::VideoParameters,
@@ -726,8 +846,7 @@ impl Packetizer {
                     };
                     // TODO: ctx, channel_id, and ssrc are placeholders.
                     return Ok(Some(Packet {
-                        ctx: crate::RtspMessageContext::dummy(),
-                        channel_id: 0,
+                        ctx: crate::PacketContext::dummy(),
                         stream_id: self.stream_id,
                         timestamp,
                         ssrc: 0,
@@ -750,8 +869,7 @@ impl Packetizer {
                     mark = false;
                 }
                 Ok(Some(Packet {
-                    ctx: crate::RtspMessageContext::dummy(),
-                    channel_id: 0,
+                    ctx: crate::PacketContext::dummy(),
                     stream_id: self.stream_id,
                     timestamp,
                     ssrc: 0,
@@ -803,8 +921,7 @@ impl Packetizer {
                 }
                 // TODO: placeholders.
                 Ok(Some(Packet {
-                    ctx: crate::RtspMessageContext::dummy(),
-                    channel_id: 0,
+                    ctx: crate::PacketContext::dummy(),
                     stream_id: self.stream_id,
                     timestamp,
                     ssrc: 0,
@@ -910,8 +1027,7 @@ mod tests {
         };
         d.push(Packet {
             // plain SEI packet.
-            ctx: crate::RtspMessageContext::dummy(),
-            channel_id: 0,
+            ctx: crate::PacketContext::dummy(),
             stream_id: 0,
             timestamp,
             ssrc: 0,
@@ -924,8 +1040,7 @@ mod tests {
         assert!(d.pull().is_none());
         d.push(Packet {
             // STAP-A packet.
-            ctx: crate::RtspMessageContext::dummy(),
-            channel_id: 0,
+            ctx: crate::PacketContext::dummy(),
             stream_id: 0,
             timestamp,
             ssrc: 0,
@@ -938,8 +1053,7 @@ mod tests {
         assert!(d.pull().is_none());
         d.push(Packet {
             // FU-A packet, start.
-            ctx: crate::RtspMessageContext::dummy(),
-            channel_id: 0,
+            ctx: crate::PacketContext::dummy(),
             stream_id: 0,
             timestamp,
             ssrc: 0,
@@ -952,8 +1066,7 @@ mod tests {
         assert!(d.pull().is_none());
         d.push(Packet {
             // FU-A packet, middle.
-            ctx: crate::RtspMessageContext::dummy(),
-            channel_id: 0,
+            ctx: crate::PacketContext::dummy(),
             stream_id: 0,
             timestamp,
             ssrc: 0,
@@ -966,8 +1079,7 @@ mod tests {
         assert!(d.pull().is_none());
         d.push(Packet {
             // FU-A packet, end.
-            ctx: crate::RtspMessageContext::dummy(),
-            channel_id: 0,
+            ctx: crate::PacketContext::dummy(),
             stream_id: 0,
             timestamp,
             ssrc: 0,
@@ -990,6 +1102,163 @@ mod tests {
         );
     }
 
+    /// Test bad framing at the start of stream from a Reolink RLC-822A
+    /// Reolink RLC-822A (IPC_523128M8MP) running firmware v3.0.0.177_21012101:
+    /// suppress incorrect access unit changes after the SPS and PPS.
+    #[test]
+    fn depacketize_reolink_bad_framing_at_start() {
+        let mut d = super::Depacketizer::new(90_000, Some("packetization-mode=1;profile-level-id=640033;sprop-parameter-sets=Z2QAM6wVFKCgL/lQ,aO48sA==")).unwrap();
+        let ts1 = crate::Timestamp {
+            timestamp: 0,
+            clock_rate: NonZeroU32::new(90_000).unwrap(),
+            start: 0,
+        };
+        let ts2 = crate::Timestamp {
+            timestamp: 1,
+            clock_rate: NonZeroU32::new(90_000).unwrap(),
+            start: 0,
+        };
+        d.push(Packet {
+            // SPS with (incorrect) mark
+            ctx: crate::PacketContext::dummy(),
+            stream_id: 0,
+            timestamp: ts1,
+            ssrc: 0,
+            sequence_number: 0,
+            loss: 0,
+            mark: true,
+            payload: Bytes::from_static(b"\x67\x64\x00\x33\xac\x15\x14\xa0\xa0\x2f\xf9\x50"),
+        })
+        .unwrap();
+        assert!(d.pull().is_none());
+        d.push(Packet {
+            // PPS
+            ctx: crate::PacketContext::dummy(),
+            stream_id: 0,
+            timestamp: ts1,
+            ssrc: 0,
+            sequence_number: 1,
+            loss: 0,
+            mark: false,
+            payload: Bytes::from_static(b"\x68\xee\x3c\xb0"),
+        })
+        .unwrap();
+        assert!(d.pull().is_none());
+        d.push(Packet {
+            // Slice layer without partitioning IDR.
+            // This has a different timestamp than the SPS and PPS, even though
+            // RFC 6184 section 5.1 says that "the timestamp must match that of
+            // the primary coded picture of the access unit and that the marker
+            // bit can only be set on the final packet of the access unit.""
+            ctx: crate::PacketContext::dummy(),
+            stream_id: 0,
+            timestamp: ts2,
+            ssrc: 0,
+            sequence_number: 2,
+            loss: 0,
+            mark: true,
+            payload: Bytes::from_static(b"\x65slice"),
+        })
+        .unwrap();
+        let frame = match d.pull() {
+            Some(CodecItem::VideoFrame(frame)) => frame,
+            o => panic!("unexpected pull result {:#?}", o),
+        };
+        assert_eq!(
+            &frame.data()[..],
+            b"\x00\x00\x00\x0C\x67\x64\x00\x33\xac\x15\x14\xa0\xa0\x2f\xf9\x50\
+              \x00\x00\x00\x04\x68\xee\x3c\xb0\
+              \x00\x00\x00\x06\x65slice"
+        );
+        assert_eq!(frame.timestamp, ts2); // use the timestamp from the video frame.
+    }
+
+    /// Test bad framing at a GOP boundary in a stream from a Reolink RLC-822A
+    /// Reolink RLC-822A (IPC_523128M8MP) running firmware v3.0.0.177_21012101:
+    /// suppress incorrect access unit changes after the SPS and PPS.
+    #[test]
+    fn depacketize_reolink_gop_boundary() {
+        let mut d = super::Depacketizer::new(90_000, Some("packetization-mode=1;profile-level-id=640033;sprop-parameter-sets=Z2QAM6wVFKCgL/lQ,aO48sA==")).unwrap();
+        let ts1 = crate::Timestamp {
+            timestamp: 0,
+            clock_rate: NonZeroU32::new(90_000).unwrap(),
+            start: 0,
+        };
+        let ts2 = crate::Timestamp {
+            timestamp: 1,
+            clock_rate: NonZeroU32::new(90_000).unwrap(),
+            start: 0,
+        };
+        d.push(Packet {
+            // Slice layer without partitioning non-IDR, representing the
+            // last frame of the previous GOP.
+            ctx: crate::PacketContext::dummy(),
+            stream_id: 0,
+            timestamp: ts1,
+            ssrc: 0,
+            sequence_number: 0,
+            loss: 0,
+            mark: true,
+            payload: Bytes::from_static(b"\x01slice"),
+        })
+        .unwrap();
+        let frame = match d.pull() {
+            Some(CodecItem::VideoFrame(frame)) => frame,
+            o => panic!("unexpected pull result {:#?}", o),
+        };
+        assert_eq!(&frame.data()[..], b"\x00\x00\x00\x06\x01slice");
+        assert_eq!(frame.timestamp, ts1);
+        d.push(Packet {
+            // SPS with (incorrect) timestamp matching last frame.
+            ctx: crate::PacketContext::dummy(),
+            stream_id: 0,
+            timestamp: ts1,
+            ssrc: 0,
+            sequence_number: 1,
+            loss: 0,
+            mark: false, // correctly has no mark, unlike first SPS in stream.
+            payload: Bytes::from_static(b"\x67\x64\x00\x33\xac\x15\x14\xa0\xa0\x2f\xf9\x50"),
+        })
+        .unwrap();
+        assert!(d.pull().is_none());
+        d.push(Packet {
+            // PPS, again with timestamp matching last frame.
+            ctx: crate::PacketContext::dummy(),
+            stream_id: 0,
+            timestamp: ts1,
+            ssrc: 0,
+            sequence_number: 2,
+            loss: 0,
+            mark: false,
+            payload: Bytes::from_static(b"\x68\xee\x3c\xb0"),
+        })
+        .unwrap();
+        assert!(d.pull().is_none());
+        d.push(Packet {
+            // Slice layer without partitioning IDR. Now correct timestamp.
+            ctx: crate::PacketContext::dummy(),
+            stream_id: 0,
+            timestamp: ts2,
+            ssrc: 0,
+            sequence_number: 3,
+            loss: 0,
+            mark: true,
+            payload: Bytes::from_static(b"\x65slice"),
+        })
+        .unwrap();
+        let frame = match d.pull() {
+            Some(CodecItem::VideoFrame(frame)) => frame,
+            o => panic!("unexpected pull result {:#?}", o),
+        };
+        assert_eq!(
+            &frame.data()[..],
+            b"\x00\x00\x00\x0C\x67\x64\x00\x33\xac\x15\x14\xa0\xa0\x2f\xf9\x50\
+              \x00\x00\x00\x04\x68\xee\x3c\xb0\
+              \x00\x00\x00\x06\x65slice"
+        );
+        assert_eq!(frame.timestamp, ts2); // use the timestamp from the video frame.
+    }
+
     #[test]
     fn depacketize_parameter_change() {
         let mut d = super::Depacketizer::new(90_000, Some("a=fmtp:96 profile-level-id=420029; packetization-mode=1; sprop-parameter-sets=Z01AHppkBYHv/lBgYGQAAA+gAAE4gBA=,aO48gA==")).unwrap();
@@ -1005,8 +1274,7 @@ mod tests {
             start: 0,
         };
         d.push(Packet { // new SPS.
-            ctx: crate::RtspMessageContext::dummy(),
-            channel_id: 0,
+            ctx: crate::PacketContext::dummy(),
             stream_id: 0,
             timestamp,
             ssrc: 0,
@@ -1018,20 +1286,32 @@ mod tests {
         assert!(d.pull().is_none());
         d.push(Packet {
             // same PPS again.
-            ctx: crate::RtspMessageContext::dummy(),
-            channel_id: 0,
+            ctx: crate::PacketContext::dummy(),
             stream_id: 0,
             timestamp,
             ssrc: 0,
             sequence_number: 1,
             loss: 0,
-            mark: true,
+            mark: false,
             payload: Bytes::from_static(b"\x68\xee\x3c\x80"),
         })
         .unwrap();
+        assert!(d.pull().is_none());
+        d.push(Packet {
+            // dummy slice NAL to end the AU.
+            ctx: crate::PacketContext::dummy(),
+            stream_id: 0,
+            timestamp,
+            ssrc: 0,
+            sequence_number: 2,
+            loss: 0,
+            mark: true,
+            payload: Bytes::from_static(b"\x65slice"),
+        })
+        .unwrap();
         let frame = match d.pull() {
             Some(CodecItem::VideoFrame(frame)) => frame,
-            _ => panic!(),
+            o => panic!("unexpected pull result {:#?}", o),
         };
         assert!(frame.new_parameters.is_some());
         let p = frame.new_parameters.unwrap();
